/*
 * Copyright 2017 Arne Vajh�j.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

package dk.vajhoej.vms.call.lib;

import static dk.vajhoej.vms.call.VMS.*;
import static dk.vajhoej.vms.call.lib.VMSCodes.*;

import java.io.UnsupportedEncodingException;

import dk.vajhoej.vms.call.VMSLibrary;
import dk.vajhoej.vms.call.VMS.CharacterString;
import dk.vajhoej.vms.call.VMS.LongWord;
import dk.vajhoej.vms.call.VMS.Word;

/**
 * Class LIBRTL is a wrapper library for LIB$ functions.
 * 
 * See VMS documentation for semantics.
 */
public class LIBRTL extends VMSLibrary {
	/**
	 * Call LIB$CREATE_DIR to create directory.
	 * @param dirnam directory name
	 * @return return status code
	 * @throws UnsupportedEncodingException if directory name can not be encoded
	 */
	public static Result<Void> LIB$CREATE_DIR(String dirnam) throws UnsupportedEncodingException {
		int sts = call("LIBRTL", "LIB$CREATE_DIR", pass(new CharacterString(dirnam)).byDescriptor().readOnly());
		return new Result<Void>(sts);
	}
	/**
	 * Call LIB$DELETE_FILE to delete file.
	 * @param fnm filename
	 * @return return status code
	 * @throws UnsupportedEncodingException if filename can not be encoded
	 */
	public static Result<Void> LIB$DELETE_FILE(String fnm) throws UnsupportedEncodingException {
		int sts = call("LIBRTL", "LIB$DELETE_FILE", pass(new CharacterString(fnm)).byDescriptor().readOnly());
		return new Result<Void>(sts);
	}
	/**
	 * Call LIB$DELETE_LOGICAL to delete logical.
	 * @param lognam logical name
	 * @return return status code
	 * @throws UnsupportedEncodingException if logical name can not be encoded
	 */
	public static Result<Void> LIB$DELETE_LOGICAL(String lognam) throws UnsupportedEncodingException {
		int sts = call("LIBRTL", "LIB$DELETE_LOGICAL", pass(new CharacterString(lognam)).byDescriptor().readOnly());
		return new Result<Void>(sts);
	}
	/**
	 * Call LIB$DELETE_SYMBOL to delete symbol.
	 * @param symnam symbol name
	 * @return return status code
	 * @throws UnsupportedEncodingException if symbol name can not be encoded
	 */
	public static Result<Void> LIB$DELETE_SYMBOL(String symnam) throws UnsupportedEncodingException {
		int sts = call("LIBRTL", "LIB$DELETE_SYMBOL", pass(new CharacterString(symnam)).byDescriptor().readOnly());
		return new Result<Void>(sts);
	}
	/**
	 * Call LIB$DISABLE_CTRL to disable CTRL/T or CTRL/Y.
	 * @param ctrlt disable CTRL/T
	 * @param ctrly disable CTRL/Y
	 * @return return status code
	 */
	public static Result<Void> LIB$DISABLE_CTRL(boolean ctrlt, boolean ctrly) {
		LongWord mask = new LongWord((ctrlt ? LIB$M_CLI_CTRLT : 0x00000000) & (ctrly ? LIB$M_CLI_CTRLY : 0x00000000));
		int sts = call("LIBRTL", "LIB$DISABLE_CTRL", pass(mask).byReference().readOnly());
		return new Result<Void>(sts);		
	}
	/**
	 * Call LIB$DO_COMMAND to transfer control to execution of command.
	 * @param cmd command to execute
	 * @return return status code
	 * @throws UnsupportedEncodingException if command can not be encoded
	 */
	public static Result<Void> LIB$DO_COMMAND(String cmd) throws UnsupportedEncodingException {
		int sts = call("LIBRTL", "LIB$DO_COMMAND", pass(new CharacterString(cmd)).byDescriptor().readOnly());
		return new Result<Void>(sts);
	}
	/**
	 * Call LIB$ENABLE_CTRL to enable CTRL/T or CTRL/Y.
	 * @param ctrlt disable CTRL/T
	 * @param ctrly disable CTRL/Y
	 * @return return status code
	 */
	public static Result<Void> LIB$ENABLE_CTRL(boolean ctrlt, boolean ctrly) {
		LongWord mask = new LongWord((ctrlt ? LIB$M_CLI_CTRLT : 0x00000000) & (ctrly ? LIB$M_CLI_CTRLY : 0x00000000));
		int sts = call("LIBRTL", "LIB$ENABLE_CTRL", pass(mask).byReference().readOnly());
		return new Result<Void>(sts);		
	}
	/**
	 * Call LIB$EXPAND_NODENAME to expand nodename.
	 * @param nodnam nodename
	 * @return return status code + expanded node ame
	 * @throws UnsupportedEncodingException if nodename can not be encoded or decoded
	 */
	public static Result<String> LIB$EXPAND_NODENAME(String nodnam) throws UnsupportedEncodingException {
        CharacterString buf = new CharacterString(255);
        Word retlen = new Word();
        int sts = call("LIBRTL", "LIB$EXPAND_NODENAME", pass(new CharacterString(nodnam)).byDescriptor().readOnly(),
                                                        pass(buf).byDescriptor().writeOnly(),
                                                        pass(retlen).byReference().writeOnly());
        String res = createString(buf, retlen);
        return new Result<String>(sts, res);
	}
	/**
	 * Class FIND_FILE_CTX represents context for LIB$FIND_FILE and LIB$FIND_FILE_END calls.
	 */
	public static class FIND_FILE_CTX {
		private String filespec;
		private LongWord ctx = new LongWord(0);
		/**
		 * Construct instance.
		 * @param filespec file specification
		 */
		public FIND_FILE_CTX(String filespec) {
			this.filespec = filespec;
		}
		/**
		 * Call LIB$FIND_FILE to get next matching file.
		 * @return return status code + name of matching file
		 * @throws UnsupportedEncodingException if file spec can not be encoded
		 */
		public Result<String> LIB$FIND_FILE() throws UnsupportedEncodingException {
	        CharacterString buf = new CharacterString(1024);
	        int sts = call("LIBRTL", "LIB$FIND_FILE", pass(new CharacterString(filespec)).byDescriptor().readOnly(),
	                                                  pass(buf).byDescriptor().readWrite(),
	                                                  pass(ctx).byReference().readWrite());
	        String res = createTrimmedString(buf);
	        return new Result<String>(sts, res);
		}
		/**
		 * Call LIB$FIND_FILE_END to end processing.
		 * @return return status code
		 */
		public Result<Void> LIB$FIND_FILE_END() {
			int sts = call("LIBRTL", "LIB$FIND_FILE_END", pass(ctx).byReference().readOnly());
			return new Result<Void>(sts);
		}
	}
	/**
	 * Call LIB$GETDVI to get single item as int.
	 * @param code item
	 * @param devnam device name
	 * @return return status code + value
	 * @throws UnsupportedEncodingException if device name can not be encoded
	 */
	public static Result<Integer> LIB$GETDVI_AsInt(int code, String devnam) throws UnsupportedEncodingException {
		LongWord res = new LongWord(0);
        int sts = call("LIBRTL", "LIB$GETDVI", pass(new LongWord(code)).byReference().readOnly(),
        		                               MISSING,
        		                               pass(new CharacterString(devnam)).byDescriptor().readOnly(),
        		                               pass(res).byReference().writeOnly());
        return new Result<Integer>(sts, res.getValue());
	}
	/**
	 * Call LIB$GETDVI to get single item as String.
	 * @param code item
	 * @param devnam device name
	 * @return return status code + value
	 * @throws UnsupportedEncodingException if device name can not be encoded or value can not be decoded
	 */
	public static Result<String> LIB$GETDVI_AsString(int code, String devnam) throws UnsupportedEncodingException {
        CharacterString buf = new CharacterString(255);
        Word retlen = new Word();
        int sts = call("LIBRTL", "LIB$GETDVI", pass(new LongWord(code)).byReference().readOnly(),
        		                               MISSING,
        		                               pass(new CharacterString(devnam)).byDescriptor().readOnly(),
        		                               MISSING,
                                               pass(buf).byDescriptor().writeOnly(),
                                               pass(retlen).byReference().writeOnly());
        String res = createString(buf, retlen);
        return new Result<String>(sts, res);
	}
	/**
	 * Call LIB$GETJPI to get single item as int.
	 * @param code item
	 * @param pid PID
	 * @return return status code + value
	 * @throws UnsupportedEncodingException will never happen
	 */
	public static Result<Integer> LIB$GETJPI_AsInt(int code, int pid) throws UnsupportedEncodingException {
		LongWord res = new LongWord(0);
        int sts = call("LIBRTL", "LIB$GETJPI", pass(new LongWord(code)).byReference().readOnly(),
        		                               pass(new LongWord(pid)).byReference().readWrite(),
        		                               MISSING,
        		                               pass(res).byReference().writeOnly());
        return new Result<Integer>(sts, res.getValue());
	}
	/**
	 * Call LIB$GETJPI to get single item as String.
	 * @param code item
	 * @param pid PID
	 * @return return status code + value
	 * @throws UnsupportedEncodingException if value can not be decoded
	 */
	public static Result<String> LIB$GETJPI_AsString(int code, int pid) throws UnsupportedEncodingException {
        CharacterString buf = new CharacterString(255);
        Word retlen = new Word();
        int sts = call("LIBRTL", "LIB$GETJPI", pass(new LongWord(code)).byReference().readOnly(),
                                               pass(new LongWord(pid)).byReference().readWrite(),
        		                               MISSING,
        		                               MISSING,
                                               pass(buf).byDescriptor().writeOnly(),
                                               pass(retlen).byReference().writeOnly());
        String res = createString(buf, retlen);
        return new Result<String>(sts, res);
	}
	/**
	 * Call LIB$GETSYI to get single item as int.
	 * @param code item
	 * @return return status code + value
	 * @throws UnsupportedEncodingException will never happen
	 */
	public static Result<Integer> LIB$GETSYI_AsInt(int code) throws UnsupportedEncodingException {
		LongWord res = new LongWord(0);
        int sts = call("LIBRTL", "LIB$GETSYI", pass(new LongWord(code)).byReference().readOnly(),
                                               pass(res).byReference().writeOnly(),
        		                               MISSING,
        		                               MISSING,
        		                               MISSING,
        		                               MISSING);
        return new Result<Integer>(sts, res.getValue());
	}
	/**
	 * Call LIB$GETSYI to get single item as String.
	 * @param code item
	 * @return return status code + value
	 * @throws UnsupportedEncodingException if value can not be decoded
	 */
	public static Result<String> LIB$GETSYI_AsString(int code) throws UnsupportedEncodingException {
        CharacterString buf = new CharacterString(255);
        Word retlen = new Word();
        int sts = call("LIBRTL", "LIB$GETSYI", pass(new LongWord(code)).byReference().readOnly(),
        		                               MISSING,
                                               pass(buf).byDescriptor().writeOnly(),
                                               pass(retlen).byReference().writeOnly(),
                                               MISSING,
                                               MISSING);
        String res = createString(buf, retlen);
        return new Result<String>(sts, res);
	}
	/**
	 * Call LIB$GET_COMMON to load string from common area.
	 * @return return status code + string value
	 * @throws UnsupportedEncodingException if string value can not be decoded
	 */
    public static Result<String> LIB$GET_COMMON() throws UnsupportedEncodingException {
        CharacterString buf = new CharacterString(255);
        Word retlen = new Word();
        int sts = call("LIBRTL", "LIB$GET_COMMON", pass(buf).byDescriptor().writeOnly(),
                                                   pass(retlen).byReference().writeOnly());
        String res = createString(buf, retlen);
        return new Result<String>(sts, res);
	}
    /**
     * Call LIB$GET_FOREIGN to get command line arguments.
     * @return return status code + command line arguments
     * @throws UnsupportedEncodingException if command line arguments can not be decoded
     */
	public static Result<String> LIB$GET_FOREIGN() throws UnsupportedEncodingException {
        CharacterString buf = new CharacterString(1024);
        Word retlen = new Word();
        int sts = call("LIBRTL", "LIB$GET_FOREIGN", pass(buf).byDescriptor().writeOnly(),
        		                                    MISSING,
                                                    pass(retlen).byReference().writeOnly());
        String res = createString(buf, retlen);
        return new Result<String>(sts, res);
	}
	/**
	 * Call LIB$GET_LOGICAL to get value of logical name.
	 * @param lognam logical name
	 * @return return status code + value of logical name
	 * @throws UnsupportedEncodingException if logical name can not be encoded or value can not be decoded
	 */
	public static Result<String> LIB$GET_LOGICAL(String lognam) throws UnsupportedEncodingException {
        CharacterString buf = new CharacterString(255);
        Word retlen = new Word();
        int sts = call("LIBRTL", "LIB$GET_LOGICAL", pass(new CharacterString(lognam)).byDescriptor().readOnly(),
                                                    pass(buf).byDescriptor().writeOnly(),
                                                    pass(retlen).byReference().writeOnly());
        String res = createString(buf, retlen);
        return new Result<String>(sts, res);
	}
	/**
	 * Call LIB$GET_SYMBOL to get value of symbol.
	 * @param symnam symbol name
	 * @return return status code + value of symbol
	 * @throws UnsupportedEncodingException if symbol name can not be encoded or value can not be decoded
	 */
    public static Result<String> LIB$GET_SYMBOL(String symnam) throws UnsupportedEncodingException {
        CharacterString buf = new CharacterString(255);
        Word retlen = new Word();
        int sts = call("LIBRTL", "LIB$GET_SYMBOL", pass(new CharacterString(symnam)).byDescriptor().readOnly(),
                                                   pass(buf).byDescriptor().writeOnly(),
                                                   pass(retlen).byReference().writeOnly());
        String res = createString(buf, retlen);
        return new Result<String>(sts, res);
	}
    /**
     * Call LIB$PAUSE to pause.
     * @return return status code
     */
    public static Result<Void> LIB$PAUSE() {
        int sts = call("LIBRTL", "LIB$PAUSE");
        return new Result<Void>(sts);
    }
    /**
     * Call LIB$PUT_COMMON to save string to common area.
     * @param value string value
     * @return return code
     * @throws UnsupportedEncodingException if string value can not be encoded
     */
    public static Result<Void> LIB$PUT_COMMON(String value) throws UnsupportedEncodingException {
        int sts = call("LIBRTL", "LIB$PUT_COMMON", pass(new CharacterString(value)).byDescriptor().readOnly());
        return new Result<Void>(sts);
    }
    /**
     * Call LIB$SET_LOGICAL to define logical.
     * @param lognam logical name
     * @param value value for logical name
     * @return return status code
     * @throws UnsupportedEncodingException if either logical name or value can not be encoded
     */
    public static Result<Void> LIB$SET_LOGICAL(String lognam, String value) throws UnsupportedEncodingException {
        int sts = call("LIBRTL", "LIB$SET_LOGICAL", pass(new CharacterString(lognam)).byDescriptor().readOnly(),
                		                            pass(new CharacterString(value)).byDescriptor().readOnly());
        return new Result<Void>(sts);
    }
    /**
     * Call LIB$SET_LOGICAL to define logical in specific logical table.
     * @param lognam logical name
     * @param value value for logical name
     * @param tblnam logical table name
     * @return return status code
     * @throws UnsupportedEncodingException if either logical name or value can not be encoded
     */
	public static Result<Void> LIB$SET_LOGICAL(String lognam, String value, String tblnam) throws UnsupportedEncodingException {
        int sts = call("LIBRTL", "LIB$SET_LOGICAL", pass(new CharacterString(lognam)).byDescriptor().readOnly(),
        		                                    pass(new CharacterString(value)).byDescriptor().readOnly(),
        		                                    pass(new CharacterString(tblnam)).byDescriptor().readOnly());
        return new Result<Void>(sts);
	}
    /**
     * Call LIB$SET_SYMBOL to set symbol.
     * @param symnam symbol name
     * @param value value for lsymbol
     * @return return status code
     * @throws UnsupportedEncodingException if either symbol name or value can not be encoded
     */
    public static Result<Void> LIB$SET_SYMBOL(String symnam, String value) throws UnsupportedEncodingException {
        int sts = call("LIBRTL", "LIB$SET_SYMBOL", pass(new CharacterString(symnam)).byDescriptor().readOnly(),
                		                           pass(new CharacterString(value)).byDescriptor().readOnly(),
                		                           pass(new LongWord(LIB$K_CLI_GLOBAL_SYM)).byReference().readOnly());
        return new Result<Void>(sts);
    }
    /**
     * Call LIB$SUB_TIMES to subtract start time from end time.
     * @param end end VMS time
     * @param start start VMS time
     * @return return status code + delta time
     */
    public static Result<Long> LIB$SUB_TIMES(long end, long start) {
    	QuadWord res = new QuadWord();
        int sts = call("LIBRTL", "LIB$SUB_TIMES", pass(new QuadWord(end)).byReference().readOnly(),
									        	  pass(new QuadWord(start)).byReference().readOnly(),
									        	  pass(res).byReference().writeOnly());
        return new Result<Long>(sts, res.getValue());
    }
    /**
     * Call LIB$SYS_ASCTIM to format a VMS time.
     * @param vmstime VMS time
     * @return return status code + formatted time
     * @throws UnsupportedEncodingException if formatted time can not be decoded
     */
    public static Result<String> LIB$SYS_ASCTIM(long vmstime) throws UnsupportedEncodingException {
        CharacterString buf = new CharacterString(255);
        Word retlen = new Word();
        int sts = call("LIBRTL", "LIB$SYS_ASCTIM", pass(retlen).byReference().writeOnly(),
        		                                   pass(buf).byDescriptor().writeOnly(),
                                                   pass(new QuadWord(vmstime)).byReference().readOnly());
        String res = createString(buf, retlen);
        return new Result<String>(sts, res);
	}
    /**
     * Call LIB$SYS_GETMSG to get message text for error code.
     * @param msgid error code
     * @return return status code + message text
     * @throws UnsupportedEncodingException if message text can not be decoded
     */
    public static Result<String> LIB$SYS_GETMSG(int msgid) throws UnsupportedEncodingException {
        CharacterString buf = new CharacterString(255);
        Word retlen = new Word();
        int sts = call("LIBRTL", "LIB$SYS_GETMSG", pass(new LongWord(msgid)).byReference().readOnly(),
        		                                   pass(retlen).byReference().writeOnly(),
        		                                   pass(buf).byDescriptor().writeOnly());
        String res = createString(buf, retlen);
        return new Result<String>(sts, res);
    }
    /**
     * Class TIMER_CTX represent context for LIB$INIT_TIMER and LIB$STAT_TIMER calls.
     */
	public static class TIMER_CTX {
		private LongWord ctx = new LongWord(0);
		/**
		 * Call LIB$INIT_TIMER to start timer.
		 * @return return status code
		 */
		public Result<Void> LIB$INIT_TIMER() {
	        int sts = call("LIBRTL", "LIB$INIT_TIMER", pass(ctx).byReference().readWrite());
	        return new Result<Void>(sts);
		}
		/**
		 * Call LIB$STAT_TIMER to get timer values.
		 * @param code code for timer value
		 * @return return status code + timer value
		 */
		public Result<Integer> LIB$STAT_TIMER(int code) {
			LongWord res = new LongWord();
			int sts = call("LIBRTL", "LIB$STAT_TIMER", pass(new LongWord(code)).byReference().readOnly(),
					                                   pass(res).byReference().writeOnly());
			return new Result<Integer>(sts, res.getValue());
		}
	}
	/**
	 * Call LIB$WAIT to sleep.
	 * @param seconds number seconds to sleep
	 * @return return status code
	 */
	public static Result<Void> LIB$WAIT(double seconds) {
		int sts = call("LIBRTL", "LIB$WAIT", pass(new TFloat(seconds)).byReference().readOnly(),
				                             MISSING,
				                             pass(new LongWord(LIB$K_IEEE_T)).byReference().readOnly());
		return new Result<Void>(sts);		
	}
}
