/*
 * Copyright 2017 Arne Vajhj.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

package dk.vajhoej.vms.call.lib;

import static dk.vajhoej.vms.call.VMS.*;
import static dk.vajhoej.vms.call.lib.VMSCodes.*;

import java.io.UnsupportedEncodingException;

import dk.vajhoej.record.ArrayField;
import dk.vajhoej.record.FieldType;
import dk.vajhoej.record.RecordException;
import dk.vajhoej.record.Struct;
import dk.vajhoej.record.StructField;
import dk.vajhoej.record.StructInfo;
import dk.vajhoej.record.StructInfoCache;
import dk.vajhoej.vms.call.VMSLibrary;

/**
 * Class RMS is a wrapper library for RMS functions.
 * 
 * See VMS documentation for semantics.
 */
public class RMS extends VMSLibrary {
	@Struct
	public static class StringBuffer {
		@StructField(n=0,type=FieldType.VARSTR,prefixlength=4)
		private String buffer;
		public StringBuffer() {
			this("");
		}
		public StringBuffer(String buffer) {
			this.buffer = buffer;
		}
		public String getBuffer() {
			return buffer;
		}
	}
	@Struct
	public static class ByteArrayBuffer {
		@ArrayField(countprefix=4)
		@StructField(n=0,type=FieldType.INT1)
		private byte[] buffer;
		public ByteArrayBuffer() {
			this(new byte[0]);
		}
		public ByteArrayBuffer(byte[] buffer) {
			this.buffer = buffer;
		}
		public byte[] getBuffer() {
			return buffer;
		}
	}
	private static int fablen;
	private static int fnaix;
	private static int rablen;
	private static int ubfix;
	private static int rbfix;
	private static int fabix;
	static {
		try {
			StructInfo fabsi = StructInfoCache.analyze(FAB.class);
			fablen = fabsi.getLength();
			fnaix = fabsi.getOffset("fna");
			StructInfo rabsi = StructInfoCache.analyze(RAB.class);
			rablen = rabsi.getLength();
			ubfix = rabsi.getOffset("ubf");
			rbfix = rabsi.getOffset("rbf");
			fabix = rabsi.getOffset("fab");
		} catch (RecordException e) {
			throw new RuntimeException("Internal problem: " + e);
		}
	}
	private static MultiBlock blockPack(FAB fab) throws RecordException {
		return new MultiBlock(new BlockHeader(fnaix, fablen + 4), // skip FAB and prefix length
				              fab,
				              new StringBuffer(fab.getFilename()));
	}
	private static MultiBlock blockPack(RAB rab) throws RecordException {
		if(rab.getUsrbuf() == null && rab.getRecbuf() == null) {
			return new MultiBlock(new BlockHeader(fabix, rablen, rablen + fnaix, rablen + fablen + 4),
		              			  rab,
		              			  rab.getFabblk(),
		              			  new StringBuffer(rab.getFabblk().getFilename()));
		} else if(rab.getUsrbuf() != null){
			return new MultiBlock(new BlockHeader(ubfix, rablen + 4, fabix, rablen + 4 + rab.getUsz(), rablen + 4 + rab.getUsz() + fnaix, rablen + 4 + rab.getUsz() + fablen + 4),
		              			  rab,
		              			  new ByteArrayBuffer(rab.getUsrbuf()),
		              			  rab.getFabblk(),
		              			  new StringBuffer(rab.getFabblk().getFilename()));
		} else if(rab.getRecbuf() != null){
			return new MultiBlock(new BlockHeader(rbfix, rablen + 4, fabix, rablen + 4 + rab.getRsz(), rablen + 4 + rab.getRsz() + fnaix, rablen + 4 + rab.getRsz() + fablen + 4),
		              			  rab,
		              			  new ByteArrayBuffer(rab.getRecbuf()),
		              			  rab.getFabblk(),
		              			  new StringBuffer(rab.getFabblk().getFilename()));
		} else {
			throw new RuntimeException("Unsupported RAB type");
		}
	}
	/**
	 * Call SYS$CLOSE to close file.
	 * @param fab FAB
	 * @return return status code
	 * @throws UnsupportedEncodingException if problem with filename and char set
	 * @throws RecordException if internal error
	 */
	public static Result<Void> SYS$CLOSE(FAB fab) throws UnsupportedEncodingException, RecordException {
		MultiBlock fabblk = blockPack(fab);
		int sts = call("SYS$PUBLIC_VECTORS", "SYS$CLOSE", pass(fabblk).byReference().readWrite());
		return new Result<Void>(sts);
	}
	/**
	 * Call SYS$CONNECT to connect to file.
	 * @param rab RAB
	 * @return return status code
	 * @throws UnsupportedEncodingException if problem with filename and char set
	 * @throws RecordException if internal error
	 */
	public static Result<Void> SYS$CONNECT(RAB rab) throws UnsupportedEncodingException, RecordException {
		MultiBlock rabblk = blockPack(rab);
		int sts = call("SYS$PUBLIC_VECTORS", "SYS$CONNECT", pass(rabblk).byReference().readWrite());
		return new Result<Void>(sts);
	}
	/**
	 * Call SYS$CREATE to create file.
	 * @param fab FAB
	 * @return return status code
	 * @throws UnsupportedEncodingException if problem with filename and char set
	 * @throws RecordException if internal error
	 */
	public static Result<Void> SYS$CREATE(FAB fab) throws UnsupportedEncodingException, RecordException {
		MultiBlock fabblk = blockPack(fab);
		int sts = call("SYS$PUBLIC_VECTORS", "SYS$CREATE", pass(fabblk).byReference().readWrite());
		return new Result<Void>(sts);
	}
	/**
	 * Call SYS$DISCONNECT to disconnect from file.
	 * @param rab RAB
	 * @return return status code
	 * @throws UnsupportedEncodingException if problem with filename and char set
	 * @throws RecordException if internal error
	 */
	public static Result<Void> SYS$DISCONNECT(RAB rab) throws UnsupportedEncodingException, RecordException {
		MultiBlock rabblk = blockPack(rab);
		int sts = call("SYS$PUBLIC_VECTORS", "SYS$DISCONNECT", pass(rabblk).byReference().readWrite());
		return new Result<Void>(sts);
	}
	/**
	 * Call SYS$DISPLAY to retrieve file attributes.
	 * @param fab FAB
	 * @return return status code
	 * @throws UnsupportedEncodingException if problem with filename and char set
	 * @throws RecordException if internal error
	 */
	public static Result<Void> SYS$DISPLAY(FAB fab) throws UnsupportedEncodingException, RecordException {
		MultiBlock fabblk = blockPack(fab);
		int sts = call("SYS$PUBLIC_VECTORS", "SYS$DISPLAY", pass(fabblk).byReference().readWrite());
		return new Result<Void>(sts);
	}
	/**
	 * Call SYS$GET to get record.
	 * @param rab RAB
	 * @return return status code + record
	 * @throws UnsupportedEncodingException if problem with record and char set
	 * @throws RecordException if internal error
	 */
	public static Result<byte[]> SYS$GET(RAB rab) throws UnsupportedEncodingException, RecordException {
		MultiBlock rabblk = blockPack(rab);
		int sts = call("SYS$PUBLIC_VECTORS", "SYS$GET", pass(rabblk).byReference().readWrite());
		// copy back usrbuf
		byte[] res = new byte[rab.getRsz()];
		System.arraycopy(rabblk.getObject(1, ByteArrayBuffer.class).getBuffer(), 0, res, 0, res.length);
		rab.setUsrbuf(res);
		//
		return new Result<byte[]>(sts, rab.getUsrbuf());
	}
	/**
	 * Call SYS$PUT to put record.
	 * @param rab RAB
	 * @return return status code
	 * @throws UnsupportedEncodingException if problem with record and char set
	 * @throws RecordException if internal error
	 */
	public static Result<Void> SYS$PUT(RAB rab) throws UnsupportedEncodingException, RecordException {
		MultiBlock rabblk = blockPack(rab);
		int sts = call("SYS$PUBLIC_VECTORS", "SYS$PUT", pass(rabblk).byReference().readWrite());
		return new Result<Void>(sts);
	}
	/**
	 * Call SYS$OPEN to open file.
	 * @param fab FAB
	 * @return return status code
	 * @throws UnsupportedEncodingException if problem with filename and char set
	 * @throws RecordException if internal error
	 */
	public static Result<Void> SYS$OPEN(FAB fab) throws UnsupportedEncodingException, RecordException {
		MultiBlock fabblk = blockPack(fab);
		int sts = call("SYS$PUBLIC_VECTORS", "SYS$OPEN", pass(fabblk).byReference().readWrite());
		return new Result<Void>(sts);
	}
	/**
	 * Class RMS_IO represents the context (FAB, RAB, NAM and XAB blocks) for RMS calls.
	 */
	public static class RMS_IO {
		private FAB fab;
		private RAB rab;
		/**
		 * Call SYS$CREATE to create file and open for write.
		 * @param fnm filename
		 * @return return status code
		 * @throws UnsupportedEncodingException if problem with filename and char set
		 * @throws RecordException if internal error
		 */
		public Result<Void> SYS$CREATE_NewSeqFileForWrite(String fnm) throws UnsupportedEncodingException, RecordException {
			return SYS$CREATE_NewSeqFileForWrite(fnm, FAB$C_VAR);
		}
		/**
		 * Call SYS$CREATE to create file and open for write.
		 * @param fnm filename
		 * @param rfm record format
		 * @return return status code
		 * @throws UnsupportedEncodingException if problem with filename and char set
		 * @throws RecordException if internal error
		 */
		public Result<Void> SYS$CREATE_NewSeqFileForWrite(String fnm, int rfm) throws UnsupportedEncodingException, RecordException {
			fab = new FAB(fnm, FAB$M_PUT, 0, 0, FAB$C_SEQ, rfm, FAB$M_CR);
			return RMS.SYS$CREATE(fab);
		}
		/**
		 * Call SYS$OPEN to open file for read.
		 * @param fnm filename
		 * @return return status code
		 * @throws UnsupportedEncodingException if problem with filename and char set
		 * @throws RecordException if internal error
		 */
		public Result<Void> SYS$OPEN_OldSeqFileForRead(String fnm) throws UnsupportedEncodingException, RecordException {
			fab = new FAB(fnm, FAB$M_GET, FAB$M_SHRGET, 0);
			return RMS.SYS$OPEN(fab);
		}
		/**
		 * Call SYS$DISPLAY to retrieve file attributes.
		 * @return return status code
		 * @throws UnsupportedEncodingException if problem with filename and char set
		 * @throws RecordException if internal error
		 */
		public Result<Void> SYS$DISPLAY() throws UnsupportedEncodingException, RecordException {
			return RMS.SYS$DISPLAY(fab);
		}
		/**
		 * Call SYS$CONNECT to connect to file.
		 * @return return status code
		 * @throws UnsupportedEncodingException if problem with filename and char set
		 * @throws RecordException if internal error
		 */
		public Result<Void> SYS$CONNECT() throws UnsupportedEncodingException, RecordException {
			rab = new RAB(fab);
			return RMS.SYS$CONNECT(rab);
		}
		/**
		 * Call SYS$GET to get record as byte array.
		 * @return return status code + record
		 * @throws UnsupportedEncodingException if problem with filename and char set
		 * @throws RecordException if internal error
		 */
		public Result<byte[]> SYS$GET_Bytes() throws UnsupportedEncodingException, RecordException {
			rab.setRac(RAB$C_SEQ);
			rab.setUsrbuf(new byte[32767]);
			return RMS.SYS$GET(rab);
		}
		/**
		 * Call SYS$GET to get record as string.
		 * @return return status code + record
		 * @throws UnsupportedEncodingException if problem with filename and char set
		 * @throws RecordException if internal error
		 */
		public Result<String> SYS$GET_String() throws UnsupportedEncodingException, RecordException {
			rab.setRac(RAB$C_SEQ);
			rab.setUsrbuf(new byte[32767]);
			Result<byte[]> tmp = RMS.SYS$GET(rab);
			return new Result<String>(tmp.getStatus(), new String(tmp.getResult(), getCharSet()));
		}
		/**
		 * Call SYS$PUT to put record of byte array.
		 * @param buf record
		 * @return return status code
		 * @throws UnsupportedEncodingException if problem with record and char set
		 * @throws RecordException if internal error
		 */
		public Result<Void> SYS$PUT_Bytes(byte[] buf) throws UnsupportedEncodingException, RecordException {
			rab.setRac(RAB$C_SEQ);
			rab.setRecbuf(buf);
			return RMS.SYS$PUT(rab);
		}
		/**
		 * Call SYS$PUT to put record of string.
		 * @param buf record
		 * @return return status code
		 * @throws UnsupportedEncodingException if problem with record and char set
		 * @throws RecordException if internal error
		 */
		public Result<Void> SYS$PUT_String(String buf) throws UnsupportedEncodingException, RecordException {
			rab.setRac(RAB$C_SEQ);
			rab.setRecbuf(buf.getBytes(getCharSet()));
			return RMS.SYS$PUT(rab);
		}
		/**
		 * Call SYS$DISCONNECT to disconnect from file.
		 * @return return status code
		 * @throws UnsupportedEncodingException if problem with filename and char set
		 * @throws RecordException if internal error
		 */
		public Result<Void> SYS$DISCONNECT() throws UnsupportedEncodingException, RecordException {
			return RMS.SYS$DISCONNECT(rab);
		}
		/**
		 * Call SYS$CLOSE to close file.
		 * @return return status code
		 * @throws UnsupportedEncodingException if problem with filename and char set
		 * @throws RecordException if internal error
		 */
		public Result<Void> SYS$CLOSE() throws UnsupportedEncodingException, RecordException {
			return RMS.SYS$CLOSE(fab);
		}
		/**
		 * Get FAB block.
		 * @return FAB
		 */
		public FAB getFAB() {
			return fab;
		}
		/**
		 * Get RAB block.
		 * @return RAB
		 */
		public RAB getRAB() {
			return rab;
		}
	}
}
